/* This is a driver for a epson graphics tablet.
 * It was based on my PCIbase driver, from
 *  http://www.bolthole.com/solaris/drivers/
 *
 *    Philip Brown
 *
 *  @(#) epsonscan.c 1.14@(#)
 */

#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/usb/usba.h>

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/stream.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>


/* This value goes in the USB "policy" info */
#define EPSON_MAX_TRANSFER 10
#define EPSON_BUFFER_SIZE 512


/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"epsonscan"	
#define DEVNODE_TYPE	DDI_NT_SERIAL /* see sunddi.h */


/**********************************************************************/

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	epson_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	epson_probe(dev_info_t *dip);
static	int	epson_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	epson_open(dev_t *dev, int oflags, int otyp, cred_t *credp);

static	int	epson_print(dev_t, char*);

static	int	epson_read(dev_t dev, struct uio* ustr, cred_t * credp);

static	int	epson_close(dev_t dev, int oflags, int otyp, cred_t *credp);
static	int	epson_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	epson_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	epson_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;

/* "char/block operations" */
static struct cb_ops	epson_cb_ops = {
	epson_open,		/* open() */
	epson_close,		/* close() */
	nodev,			/* strategy()		*/
	epson_print,		/* print routine	*/
	nodev,			/* no dump routine	*/
	epson_read,		/* read() */
	epson_write,		/* write() */
	epson_ioctl,			/* generic ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP |D_64BIT,	/* safe for multi-thread/multi-processor */
				/* also theoretically 64bit safe */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops epson_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	epson_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	epson_probe,		/* devo_probe */
	epson_attach,		/* devo_attach */
	epson_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&epson_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"epson-scanner driver",
	&epson_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   epson_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * epson_glob_soft_handle;

/* defines used on the "openstate" var in softstate*/
#define EPSON_OPEN_STARTUP	0x01 /* in middle of open() */
#define EPSON_OPEN_HELD		0x02
#define EPSON_OPEN_SHUTDOWN	0x03 /* open, but transitioning to close */

/* This gets filled out on a per-instance basis by epson_attach */
struct epson_softstate {
	dev_info_t *dip; /* you always need this cached */
	kmutex_t instance_mutex,read_mutex;
	kcondvar_t read_cv,write_cv;
	
	usb_config_descr_t usb_config;

	/* NOTE: there IS NO endpoint descritor for default/control pipe */
	usb_endpoint_descr_t in_endpoint;
	usb_endpoint_descr_t out_endpoint;

	size_t max_transfer_size;

	int usb_interfacenum;
	usb_interface_descr_t usb_interface_d;
	int openstate; /* tell if we have been open()ed or not */
	usb_pipe_handle_t in_pipe_handle,out_pipe_handle;

	mblk_t *read_blk; /* bulk callback sets this */
	int dev_busy; /* because we cant hold mutexes across usb_ calls */

};


#ifdef YOUFILLTHISOUT
/* You fill out this struct to match the offsets in your device registers.
 * The "epson_regs" pointer in epson_softstate, above,  gets mapped
 * to your first set of device registers, in epson_attach().
 *
 * BE CAREFUL!! WHen initially filling this out, you probably will want
 * to do a sizeof(struct epson_regs) to see whether your compiler has
 * messed things up with padding
 *
 * Remove the ifdef after you've really filled this out.
 * Or just use byte offsets, and re-type struct epson_regs to void*
 * in the softstate struct.
 */
struct epson_regs {
	uint16_t somereg;
	uint32_t longerreg;
	
};
#endif

/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
epson_debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;
	if(debug_level==0){
		return;
	}

	va_start(args);
	fmt = va_arg(args, char *);
	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}



/*Like the epson_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See epson_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	epson_debugprint("epson: _init called\n");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&epson_global_mutex, "epson: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct epson_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&epson_glob_soft_handle,
				sizeof(struct epson_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&epson_global_mutex);
		ddi_soft_state_fini(&epson_glob_soft_handle);
		cmn_err(CE_WARN, "epson: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	epson_debugprint("epson: _init returning SUCCESS, supposedly\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	epson_debugprint("epson: _info called");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	epson_debugprint("epson: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&epson_glob_soft_handle);
		mutex_destroy(&epson_global_mutex);
	}
	return (ret);
}

/************************************************************/
/*                   struct uio  utility section            */
/************************************************************/


/* used by epson_read()  and epson_write() to figure out how many bytes of data
 * the user program wants right now
 */
static int count_uio_size(struct uio* uiostruct){
	int datalen=0;
	int vec_count;
	if(uiostruct==NULL){
		return 0;
	}
	for(vec_count=0;vec_count<uiostruct->uio_iovcnt;vec_count++){
		datalen+=uiostruct->uio_iov[vec_count].iov_len;
	}

	return datalen;
}

/* Utility function to copy write data from "normal" uio struct, to
 * STREAMS-based mblk struct.
 * We use uiomove to do so.
 * Which means THIS CANNOT BE CALLED FROM INTERRUPT CONTEXT. Sigh.
 *
 * Returns 0 on OK.
 */
static int
copy_uio_to_mblk(struct uio*uiop, mblk_t *mblk, size_t len){
	unsigned char *wptr;

	if((uiop==NULL) || (mblk==NULL)){
		epson_debugprint("INTERNAL ERROR: copy_uio_to_mblk pass NULL\n");
		return 1;
	}
	wptr=mblk->b_wptr;
	if(wptr==NULL){
		epson_debugprint("INTERNAL ERROR: b_wptr==NULL\n");
		return 1;
	}
	if(uiomove(wptr, len, UIO_WRITE, uiop) ==0){
		wptr+=len;
		mblk->b_wptr=wptr;
	} else {
		cmn_err(CE_WARN,"uiomove in copy_uio_to_mblk failed!");
		return 1;
	}

	return 0;
}

/* Utility function to copy write data from STREAMS-style mblk struct to
 * "normal" uio struct.
 * We use uiomove to do so.
 * Which means THIS CANNOT BE CALLED FROM INTERRUPT CONTEXT. Sigh.
 *
 * Normally, this is called as a result of a receive_bulk_data(),
 * so the amount of data available should be exactly equal to
 * the amount of data requested in 'len'
 * Similarly, it is assumes there is only ONE data block in mblk.
 *
 * Returns 0 on OK.
 *  Does NOT set uio_resid on uiop.
 *  DOES adjust mblk->b_rptr
 */
static int
copy_mblk_to_uio(mblk_t *mblk, struct uio*uiop, size_t len){
	unsigned char *rptr,*wptr;
	int srcdatalen;

	rptr=mblk->b_rptr;
	wptr=mblk->b_wptr;;
	srcdatalen=wptr-rptr;

	if(srcdatalen < len){
		epson_debugprint("INTERNAL ERROR: copy_mblk_to_uio insuficcient data\n");
		return 1;
	}

	if(srcdatalen > len){
		epson_debugprint("WARNING: copy_mblk_to_uio has extra data\n");
	}

	if((uiop==NULL) || (mblk==NULL)){
		epson_debugprint("INTERNAL ERROR: copy_mblk_to_uio pass NULL\n");
		return 1;
	}

	if(uiomove(rptr, len, UIO_READ, uiop) ==0){
		/*rptr+=len;
		mblk->b_rptr=rptr;
		*/
	} else {
		cmn_err(CE_WARN,"uiomove in copy_uio_to_mblk failed!");
		epson_debugprint("  uiomove info: len=%d, mblkcont=%p\n",
				 len,
				 mblk->b_cont);
		return 1;
	}

	return 0;
}



/**********************************************************************/

/* convenience routine to use before any I/O.
 * See if pipe is stick in err state, and reset if so.
 * Should only use from non-interrupt context.
 */
static void validate_pipe(usb_pipe_handle_t pipe){
	usb_pipe_state_t pipestate;
	if(usb_pipe_get_state(pipe,&pipestate,USB_FLAGS_SLEEP) == USB_SUCCESS){
		switch(pipestate){
		    case USB_PIPE_STATE_ERROR:
			epson_debugprint(" validate_pipe: ERROR. Resetting\n");
			if(usb_pipe_reset(pipe, USB_FLAGS_SLEEP,NULL,NULL) !=
			   USB_SUCCESS){
				epson_debugprint("pipe_reset FAILED??!\n");
			 } else {
				epson_debugprint("  pipe_reset succeeded\n");
			 }
			 break;
		    case USB_PIPE_STATE_IDLE:
			epson_debugprint(" epson_pipe_callback: pipe IDLE \n");
			break;
		    case USB_PIPE_STATE_ACTIVE:
			epson_debugprint(" epson_pipe_callback: pipe ACTIVE \n");
			break;
		    case USB_PIPE_STATE_SYNC_CLOSING:
			epson_debugprint(" epson_pipe_callback: pipe SYNC_CLOSING \n");
			break;
		    case USB_PIPE_STATE_ASYNC_CLOSING:
			epson_debugprint(" epson_pipe_callback: pipe ASYNC_CLOSING \n");
			break;
		    default:
			epson_debugprint(" epson_pipe_callback: pipe UNKNOWN state \n");
			break;
			
		}
	} else {
		epson_debugprint(" validate_pipe: get_state failed??\n");
	}

}
/* This is probably interrupt context. not sure.
 * mblk_t is a STREAMS Message Block. Arrg.
 * WARNING: This callback is called for BOTH in and out pipes, due to a
 * shared policy struct being used for the usb_pipe_opens.
 *
 * Im not sure when I should return anything but USB_SUCCESS.
 */
static int
epson_pipe_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,mblk_t*mblkptr)
{
	struct epson_softstate *softstate=(struct epson_softstate*)arg;
	
	epson_debugprint("epson_pipe_callback called\n");

	if(softstate== NULL){
		epson_debugprint("epson_pipe_callback: ERR: no softstate\n");
		return USB_SUCCESS;
	}
	

	if(pipe==softstate->out_pipe_handle){
		if(debug_level>1)
			epson_debugprint("epson_pipe_callback  has OUT pipe\n");

		if(mblkptr!=NULL){
			printf("pipe_callback(out): mblk!=NULL??\n");
			freemsg(mblkptr);
		}

		/* In theory, we could set some kind of
		 * "You had an error on write" flag here. 
		 */
		mutex_enter(&softstate->instance_mutex);
		epson_debugprint("epson_pipe_callback doing cv_signal for write\n");
		cv_signal(&softstate->write_cv);
		mutex_exit(&softstate->instance_mutex);

		return (USB_SUCCESS);
	}

	if(pipe!=softstate->in_pipe_handle){
		epson_debugprint("epson_pipe_callback has UNRECOGNIZED pipe\n");
		return (USB_SUCCESS);
	}

	/* ***************************************************** */
	/* The rest of this function handles bulk-read data only */
	/* ***************************************************** */
	if(debug_level>1)
		epson_debugprint("epson_pipe_callback has IN pipe\n");

	if(mblkptr ==NULL){
		printf("pipe_callback(in): mblk==NULL. nothing to do\n");
		return (USB_SUCCESS);
	}

	/* Now need to put the data in the softstate queue, so the
	 * user-level routine can use uiomove() for it.
	 */
	mutex_enter(&softstate->instance_mutex);

	if(softstate->read_blk!=NULL){
		/* Huh? Already set? should never happen! */
 		printf("epson_pipe_callback: read_blk!=NULL");

		freemsg(mblkptr);
		mutex_exit(&softstate->instance_mutex);
		return (USB_SUCCESS);
		
	}
	softstate->read_blk=mblkptr;
	cv_signal(&softstate->read_cv);

	mutex_exit(&softstate->instance_mutex);

	return (USB_SUCCESS);
}

#ifdef USING_ISOC_PIPE
/* Temp test routine, that is NOT standard callback, but
 * pipe_open callback, for now.
 */
static void
epson_test_callback(usb_opaque_t arg,int arg1, uint_t arg2)
{

	epson_debugprint("epson_TEST_callback called\n");
	return;
}
#endif /* USING_ISOC_PIPE */

/* This gets called for partial USB transfers.
 * EG: some error occurred. */
static int
epson_exception_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,
			 uint_t complete_reason,mblk_t*mblkptr,uint_t flag)
{
	int resetstat;
	struct epson_softstate *softstate=(struct epson_softstate*)arg;

	epson_debugprint("epson_EXCEPTION_callback called: reason %d\n",
			 complete_reason);
	if(pipe==softstate->in_pipe_handle){
		epson_debugprint("epson_EXCEPTION is on IN_pipe\n");
	}
	if(pipe==softstate->out_pipe_handle){
		epson_debugprint("epson_EXCEPTION is on OUT_pipe\n");
	}

	if(mblkptr){
		freemsg(mblkptr);
	}
	if(arg==NULL){
		epson_debugprint("epson_EXCEPTION_callback: arg==NULL!");
		return (USB_FAILURE);
	}
	
	mutex_enter(&softstate->instance_mutex);
	resetstat=usb_pipe_reset(pipe, USB_FLAGS_SLEEP,NULL,NULL);
	mutex_exit(&softstate->instance_mutex);

	if(resetstat==USB_SUCCESS){
		return (USB_SUCCESS);
	}
	epson_debugprint("exception_callback: usb_pipe_reset FAILED?!!");
	return (USB_FAILURE);
}

/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS
 */
static int epson_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "base-print(%d): %s",instance,string);
	return 0;
}

static int epson_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	epson_debugprint("epson: getinfo called\n");
	return DDI_FAILURE;
}




/* you probably wont even need to have a probe for a PCI driver 
 * But just in case you want one to tell you what major and minor
 * you're being probed with, here's one
 */
 
static int epson_probe(dev_info_t *dip){
#ifndef DO_PROBE_ROUTINE
   	epson_debugprint("epson: probe called\n");
#else

  	target=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"target", -1);
  	lun=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"lun", -1);
   
	epson_debugprint("probe- called fortarget=%d, lun=%d\n", target,lun);

	/* Normally, we would check the dev_info info,
	 * to see which device ID is being probed, and then
	 * check to see if the durn thing is actually there.
	 * But for now, assume .conf file is correct
	 */
   
	/* our two options for success are:
	 * DDI_PROBE_DONTCARE
	 * DDI_PROBE_SUCCESS
	 */

#endif /* DO_PROBE_ROUTINE */

	return DDI_PROBE_DONTCARE;
	/*return DDI_PROBE_FAILURE;*/
}

/* This routine can be called by epson_attach (if errors occur)
 * and epson_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void epson_cleanup(struct epson_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: epson_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	if(softstate->dip != NULL){
		mutex_destroy(&softstate->instance_mutex);
		mutex_destroy(&softstate->read_mutex);
		cv_destroy(&softstate->read_cv);
		cv_destroy(&softstate->write_cv);
	}

	ddi_soft_state_free(epson_glob_soft_handle,instance);
}


/* WARNING: This only seems to work from _attach().
 * calling it from open(), LOCKS THE SYSTEM!
 */
static void reset_scanner(struct epson_softstate *softstate){
	usb_pipe_handle_t pipe_handle;
	uint_t failreason;

	if(usb_pipe_open(softstate->dip,
			 NULL,NULL,USB_FLAGS_OPEN_EXCL,
			 &pipe_handle) != USB_SUCCESS){
		cmn_err(CE_WARN,"reset_scanner: open control pipe failed");

	}

	epson_debugprint("endpoint num of defaultpipe is %d\n",
			 usb_endpoint_num(pipe_handle));

	if(usb_pipe_sync_device_ctrl_send(pipe_handle,
			USB_DEV_REQ_RECIPIENT_ENDPOINT,
			USB_REQ_CLEAR_FEATURE,
			USB_ENDPOINT_HALT,
			usb_endpoint_num(pipe_handle),
			0, /*wLength*/
			NULL,
			&failreason,
			USB_FLAGS_ENQUEUE) != USB_SUCCESS){
		epson_debugprint("reset_scanner: USB_CLEAR_FEATURE failed\n");
		epson_debugprint("  failreason=%d\n",failreason);

	} else {
		epson_debugprint("reset_scanner: usb_clear_feature SUCCEEDED!\n");
	}

	usb_pipe_close(&pipe_handle,USB_FLAGS_SLEEP,NULL,NULL);
	
}

/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in epson_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  epson_detach()
 */
static int epson_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	struct epson_softstate *softstate;
	size_t usb_raw_len;
	uchar_t *usb_rawconf;
	usb_endpoint_descr_t endpoint0,endpoint1;
	int endpointsstat=0; /* keep track if we have assigned both in and out*/

	/* default debug level to **1**
	 * You can override in the .conf file with "debug-level=0".
	 * When you have a production-ready driver, change
	 * default to 0
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 1);

	epson_debugprint("epson: epson_attach called: debug-level=%d\n",
			 debug_level);

   	switch(cmd){
	 case DDI_RESUME:
		cmn_err(CE_NOTE, "epson: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "epson: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "epson: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	epson_debugprint("epson: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(epson_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "epson: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(epson_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "epson: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(epson_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;
	softstate->openstate = 0; /* redundant, but what the heck */
	


	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);
	mutex_init(&softstate->read_mutex, NULL, MUTEX_DRIVER, NULL);

	cv_init(&softstate->read_cv, NULL, CV_DRIVER,NULL);
	cv_init(&softstate->write_cv, NULL, CV_DRIVER,NULL);

	/* the buffer pointers get cleared in epson_open() */


	usb_rawconf = usb_get_raw_config_data(dip,&usb_raw_len);

	/* Fill out "configuration" descriptor */
	if(usb_parse_configuration_descr(usb_rawconf, usb_raw_len,
					&softstate->usb_config,
					USB_CONF_DESCR_SIZE) !=
						USB_CONF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_configuration_desc failed\n");
		epson_cleanup(softstate);
		return DDI_FAILURE;
	}
	epson_debugprint("config num is %d\n",
			 softstate->usb_config.bConfigurationValue);
	softstate->usb_interfacenum=usb_get_interface_number(dip);
	epson_debugprint("USB interface num is %d\n",
				softstate->usb_interfacenum);

	/* If interfacenum==-1, set it to 0: we are responsible for full dev*/
	if(softstate->usb_interfacenum==-1){
		softstate->usb_interfacenum=0;
	}

	/* Fill out "interface" descriptor */
	if(usb_parse_interface_descr(usb_rawconf,usb_raw_len,
				    softstate->usb_interfacenum,
				    0,
				    &softstate->usb_interface_d,
				    USB_IF_DESCR_SIZE) != USB_IF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_interface_desc failed");
		epson_cleanup(softstate);
		return DDI_FAILURE;
	}

	usb_pipe_bulk_transfer_size(dip,&softstate->max_transfer_size);
	epson_debugprint("max bulk transfer size =%u bytes \n",
	                  softstate->max_transfer_size);


	
	/* usb_interface_descr says "refer to USB 1.0/9.6.3" */
	/* number EXCLUDES the "default endpoint", aka control endpoint*/
	epson_debugprint("Number of endpoints is %d\n",
			 softstate->usb_interface_d.bNumEndpoints);

	/* usb_endpoint_descr says "refer to USB 1.0/9.6.4" */
	/* Fill out "endpoint" descriptor(s) */
	/* "parse" should say "get" */
	if(usb_parse_endpoint_descr(usb_rawconf, usb_raw_len,
					softstate->usb_interfacenum,
					0,/* alt interfacenum*/
					0,/*endpoint index num*/
					&endpoint0,
					USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_in_endpoint_descr 0 failed\n");
		epson_cleanup(softstate);
		return DDI_FAILURE;
	}

	if(usb_parse_endpoint_descr(usb_rawconf, usb_raw_len,
					softstate->usb_interfacenum,
					0,/* alt interfacenum*/
					1,/*endpoint index num*/
					&endpoint1,
					USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_out_endpoint_descr 1 failed\n");
		epson_cleanup(softstate);
		return DDI_FAILURE;
	}

	if(((endpoint0.bmAttributes & USB_EPT_ATTR_MASK) != USB_EPT_ATTR_BULK)||
	   ((endpoint1.bmAttributes & USB_EPT_ATTR_MASK) != USB_EPT_ATTR_BULK))
	{
		cmn_err(CE_WARN,"endpoints not 'BULK' type");
		epson_cleanup(softstate);
		return DDI_FAILURE;
	}

	/* Some scanners have 'in' first, some have 'out' first. Sigh.
	 * So, have to be flexible.
	 */
	switch(endpoint0.bEndpointAddress &USB_EPT_DIR_MASK){
	    case USB_EPT_DIR_IN:
		bcopy(&endpoint0,&softstate->in_endpoint,sizeof(endpoint0));
		endpointsstat|=1;
		break;
	    case USB_EPT_DIR_OUT:
		bcopy(&endpoint0,&softstate->out_endpoint,sizeof(endpoint0));
		endpointsstat|=2;
		break;
	    default:
		cmn_err(CE_WARN,"impossible error in endpoint analysis");
		epson_cleanup(softstate);
		return DDI_FAILURE;
	}
	switch(endpoint1.bEndpointAddress &USB_EPT_DIR_MASK){
	    case USB_EPT_DIR_IN:
		bcopy(&endpoint1,&softstate->in_endpoint,sizeof(endpoint1));
		endpointsstat|=1;
		break;
	    case USB_EPT_DIR_OUT:
		bcopy(&endpoint1,&softstate->out_endpoint,sizeof(endpoint1));
		endpointsstat|=2;
		break;
	    default:
		cmn_err(CE_WARN,"impossible error in endpoint analysis");
		epson_cleanup(softstate);
		return DDI_FAILURE;
	}

	if(endpointsstat != 3){
		cmn_err(CE_WARN,
		  "Paranoia pays off: we did not find both an in and out endpoint");
		epson_cleanup(softstate);
		return DDI_FAILURE;
	}

	epson_debugprint("in endpoint: num=%d(0x%x),attributes=%d, \n",
			 softstate->in_endpoint.bEndpointAddress,
			 softstate->in_endpoint.bEndpointAddress,
			 softstate->in_endpoint.bmAttributes);

	epson_debugprint("out endpoint: num=%d(0x%x),attributes=%d, \n",
			 softstate->out_endpoint.bEndpointAddress,
			 softstate->out_endpoint.bEndpointAddress,
			 softstate->out_endpoint.bmAttributes);

	epson_debugprint("INTERFACE class=%x, subclass=%x\n",
			 softstate->usb_interface_d.bInterfaceClass,
 			 softstate->usb_interface_d.bInterfaceSubClass);


	reset_scanner(softstate);

	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/XXXXXX
	 *
	 * To get an automatic matching link in /dev/XXXX, you need to 
	 * add your own line to /etc/devlink.tab
	 */
	 
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);

	/* There are no PCI-style interrupts, since we dont get to directly
	 * connect to a device.
	 * but an interrupt handler of sorts gets set up in open()
	 */



	ddi_report_dev(dip);

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}


/* return 0 on success, or errno */
static int epson_read(dev_t dev, struct uio* uiop, cred_t * credp){
	int instance=DEVT_TO_INST(dev);
	struct epson_softstate *softstate;
	mblk_t *mblkptr;
	int bytesrequested,cvstat;

	bytesrequested=count_uio_size(uiop); /* how many bytes do they WANT?*/
	epson_debugprint("epson: read called with readsize=%d\n",bytesrequested);

	softstate=ddi_get_soft_state(epson_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"epson_read failed: cannot get softstate?");
		uiop->uio_resid=bytesrequested;
		return ENXIO;
	}
	epson_debugprint("  openstate=%d\n",softstate->openstate);

	
	epson_debugprint("epson_read going for read_mutex\n");
	mutex_enter(&softstate->read_mutex);
	epson_debugprint("epson_read acquired read_mutex\n");
	if(softstate->dev_busy!=0){
		epson_debugprint("epson_read: we are already reading something\n");
		mutex_exit(&softstate->read_mutex);
		return EBUSY;
	}
	softstate->dev_busy=1;
	mutex_exit(&softstate->read_mutex);
	epson_debugprint("epson_read  released read_mutex\n");

	validate_pipe(softstate->out_pipe_handle);

	/* Kinda icky loop.
	 * We limit ourselves to only reading max_transfer_size
	 * at a time.
	 * So, copy one chunck at a time until either error,
	 * or read request is fulfuilled
	 */
	while(bytesrequested>0){
		int readbytes=bytesrequested;
		clock_t currtime;
		clock_t timeout=drv_usectohz(10000000); /* 10 seconds */

		if(readbytes > softstate->max_transfer_size){
			readbytes = softstate->max_transfer_size;
		}
		/* timeout is in seconds  */
		if(usb_pipe_receive_bulk_data(softstate->in_pipe_handle,
				      readbytes,
				      100, /* timeout */
				      USB_FLAGS_SLEEP) != USB_SUCCESS)
		{
		
			epson_debugprint("epson_read: receive_bulk_data FAILED\n");
			uiop->uio_resid=bytesrequested;
			softstate->dev_busy=0;
			return EIO;
		
		}
		/* now hopefully the bulk_read callback will trigger
		 * and then we can copy out some data to the user.
		 * Note that uiomove cannot be called from interrupt context.
		 */
		if(debug_level>1)
			epson_debugprint("epson_read: doing cv_wait_sig\n");

		currtime=ddi_get_lbolt();

		/* need this mutex both for cv holding, and because
		 * we snag data from softstate->read_blk
		 */

		mutex_enter(&softstate->instance_mutex);
		cvstat=cv_timedwait_sig(&softstate->read_cv,
					&softstate->instance_mutex,
					currtime + timeout);
		mblkptr=softstate->read_blk;
		softstate->read_blk=NULL;
		mutex_exit(&softstate->instance_mutex);

		if(debug_level>1)
			epson_debugprint("epson_read: cv_timedwait_sig returned...\n");

		if(mblkptr==NULL){
			epson_debugprint("ERROR: read_blk was NULL??\n");
			uiop->uio_resid=bytesrequested;
			softstate->dev_busy=0;
			return (ENXIO);
		}


		if(cvstat <1){
			switch(cvstat){
				case -1:
				epson_debugprint("TIMEOUT on read\n");
				break;
				case 0:
				epson_debugprint("signal interrupted read\n");
				break;
				default:
				epson_debugprint("UNKNOWN cvstat %d\n",cvstat);
				
			}
			freemsg(mblkptr);
			uiop->uio_resid=bytesrequested;
			softstate->dev_busy=0;
			return (ENXIO);
		}

		if(softstate->openstate == EPSON_OPEN_SHUTDOWN)
		{
			epson_debugprint("epson_read: shutdown detected\n");
			epson_debugprint("  openstate=%d\n",
			                softstate->openstate);
			uiop->uio_resid=bytesrequested;
			softstate->dev_busy=0;
			return (ENXIO);
		}

		
		if(copy_mblk_to_uio(mblkptr,uiop,readbytes)!=0){
			cmn_err(CE_WARN,"epson_read: ERROR from copy_mblk_to_uio");
			softstate->dev_busy=0;
			return EFAULT;
		}
		if(debug_level>1)
			epson_debugprint("  successfully copied %d bytes to user\n",
				 readbytes);
	
		bytesrequested -= readbytes;
		/* and if neccessary, loop round and read again */

	} /* while (bytesrequested > 0)  */


	softstate->dev_busy=0;
	epson_debugprint("epson_read: returning OK\n");
	return 0;
}

/* XXX in theory, should wait until device is not busy, i think?
 * (eg: not in the middle of doing a read)
 */
static int epson_write(dev_t dev, struct uio *uiop, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct epson_softstate *softstate;
	int flags=USB_FLAGS_ENQUEUE;
	int writesize, status,cvstatus;
	mblk_t *msgptr;
	clock_t currtime;
	clock_t timeout=drv_usectohz(10000000); /* 10 seconds */

	writesize=count_uio_size(uiop);

   	epson_debugprint("epson_write called with %d bytes\n",writesize);

	softstate=ddi_get_soft_state(epson_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"epson_read failed: cannot get softstate?");
		return DDI_FAILURE;
	}
	if(softstate->dev_busy){
		epson_debugprint("epson_write: returning EBUSY\n");
		return EBUSY;
	}

	ASSERT(writesize>=0);
	if(writesize==0){
		/* In theory, write() can be zero length, I suppose  
		 * but we're not going to try to actually write it to
		 * the device
		 */
		return 0;
	}

	msgptr=allocb(writesize,0);
	if(msgptr==NULL){
		cmn_err(CE_WARN,"allocb failed!");
		return ENOMEM;
	}
	if(copy_uio_to_mblk(uiop,msgptr,writesize) != 0){
		epson_debugprint("copy_uio_to_mblk returned fail\n");
		return EFAULT;
	}


	validate_pipe(softstate->out_pipe_handle);
	epson_debugprint("  epson_write: Sending %d bytes of data\n",writesize);

	status=usb_pipe_send_bulk_data(softstate->out_pipe_handle,msgptr,
				       0,flags);

	epson_debugprint("  epson_write: Waiting for cv_signal\n");

	/* Not sure why I'm doing this mutex here. Oh well.
	 * remember, this thing actually exits, then reenters the mutex.
	 * wierd.
	 *
	 */
	mutex_enter(&softstate->instance_mutex);
	currtime=ddi_get_lbolt();
	cvstatus=cv_timedwait_sig(&softstate->write_cv,
	                     &softstate->instance_mutex,
			     currtime + timeout);

	epson_debugprint("  epson_write: got cv_signal\n");
	
	mutex_exit(&softstate->instance_mutex);

	if(cvstatus<=0){
		epson_debugprint("epson_write: timeout/abort on cv_wait\n");
		uiop->uio_resid=writesize;
		return EIO;
	}

	if(status!=USB_SUCCESS){
		epson_debugprint("usb_pipe_send_bulk_data: FAILED.\n");
		/* I THINK that usb_p_s_b_d should have freed the msgptr */
		uiop->uio_resid=writesize;
		return EIO;
	}
	

	epson_debugprint("epson_write: returning successfully\n");
	uiop->uio_resid=0;
	return 0;

}


/* Note:
 * 1. We dont open any "pipes" to the device, until this point. If no-one has
 *    us open, there's no point in talking to the dev at all
 * 2. We dont use a control pipe. Everything to the epson scanners
 *    I know about is either bulk_in, or bulk_out
 */
static int epson_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	usb_pipe_policy_t policy,*policyptr;
	usb_pipe_state_t pipestate;
	struct epson_softstate *softstate;
	int instance=DEVT_TO_INST(*dev);

	epson_debugprint("epson: open for instance %d\n", instance);

	softstate=ddi_get_soft_state(epson_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is technically NOT an error; it is expected.
		 * We return ENXIO to support "deferred attach"
		 */
		return (ENXIO);
	}

#ifdef DO_PWRLVL
	epson_debugprint("epson_open: raising power level\n");
	usb_set_device_pwrlvl3(softstate->dip);
#endif

		
/* This seems to lock things up. so guess we cant reset here.
	epson_debugprint("epson_open: doing reset\n");
	reset_scanner(softstate);
	epson_debugprint("epson_open: reset done\n");
*/

	policyptr=&policy;
	policyptr->pp_version=USB_PIPE_POLICY_V_0;
	policyptr->pp_timeout_value=USB_PIPE_TIMEOUT;
	policyptr->pp_callback_arg=(void*)softstate;
	policyptr->pp_callback=epson_pipe_callback;
	policyptr->pp_exception_callback=epson_exception_callback;

	policyptr->pp_periodic_max_transfer_size=EPSON_MAX_TRANSFER;
	/* there is also pp_isoc_{exception_}callback, but
	 * we dont need them, maybe?
	 */

	mutex_enter(&softstate->instance_mutex);
	if(softstate->openstate!=0){
		epson_debugprint("epson: device already opened\n");
		mutex_exit(&softstate->instance_mutex);
		return (EBUSY);
	}
	softstate->openstate=EPSON_OPEN_STARTUP;
	mutex_exit(&softstate->instance_mutex);

	if(usb_pipe_open(softstate->dip,&softstate->in_endpoint, policyptr,
			 USB_FLAGS_SLEEP|USB_FLAGS_OPEN_EXCL,
			 &softstate->in_pipe_handle) !=USB_SUCCESS){
		cmn_err(CE_WARN,"usb_pipe_open(in) failed");
		softstate->openstate=0;
		return (ENXIO);
	}
	usb_pipe_get_state(softstate->in_pipe_handle,&pipestate,USB_FLAGS_SLEEP);
	epson_debugprint("state of in-pipe is %d\n",pipestate);
			 
	if(usb_pipe_open(softstate->dip,&softstate->out_endpoint, policyptr,
			 USB_FLAGS_SLEEP|USB_FLAGS_OPEN_EXCL,
			 &softstate->out_pipe_handle) !=USB_SUCCESS){
		cmn_err(CE_WARN,"usb_pipe_open(out) failed");
		usb_pipe_close(&softstate->in_pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL);
		softstate->openstate=0;
		return (ENXIO);
	}
	usb_pipe_get_state(softstate->out_pipe_handle,&pipestate,USB_FLAGS_SLEEP);
	epson_debugprint("state of out-pipe is %d\n",pipestate);



#ifdef USING_ISOC_PIPE
	/* This is not needed for BULK type pipes. 
	 * We explicitly call usb_pipe_receive_bulk_data instead.
	 */
	if(usb_pipe_start_polling(softstate->in_pipe_handle,
				  USB_FLAGS_SLEEP,
				  epson_test_callback,NULL) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"start_polling on IN pipe FAILED");
		softstate->openstate=0;
		return ENXIO;
	}
	epson_debugprint("Started polling on IN pipe\n");
#endif /* USING_ISOC_PIPE */

	epson_debugprint("epson_open: returning OK\n");
	softstate->openstate=EPSON_OPEN_HELD;
	epson_debugprint("   openstate=%d\n",softstate->openstate);

   	return 0;
}

static int epson_close(dev_t dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct epson_softstate *softstate;


	softstate=ddi_get_soft_state(epson_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"epson_close failed: cannot get softstate?");
		return DDI_FAILURE;
	}

   	epson_debugprint("epson: close called\n");
	mutex_enter(&softstate->instance_mutex);
	if(softstate->openstate!=EPSON_OPEN_HELD){
		cmn_err(CE_WARN,"ERROR: epson_close: we are not open");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
	}

	/* This is a little worrying doing this without mutex.
	 * but I believe the system will take care of non-reentrancy.
	 * I hope.
	 * We cant hold mutexes across usb calls. glarg.
	 */
	 
	softstate->openstate=EPSON_OPEN_SHUTDOWN;
	mutex_exit(&softstate->instance_mutex);


	usb_pipe_stop_polling(softstate->in_pipe_handle,
			      USB_FLAGS_SLEEP,NULL,NULL);
	epson_debugprint("Stopped polling on IN pipe\n");

	if(usb_pipe_close(&softstate->in_pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL ) != USB_SUCCESS)
	{
		epson_debugprint("ERROR: usb_pipe_close(in) failed?");
		return ENXIO;
	}
	softstate->in_pipe_handle=NULL;
	epson_debugprint("Closed IN pipe\n");


	if(usb_pipe_close(&softstate->out_pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL ) != USB_SUCCESS)
	{
		epson_debugprint("ERROR: usb_pipe_close(out) failed?");
		return ENXIO;
	}
	softstate->out_pipe_handle=NULL;
	epson_debugprint("Closed OUT pipe \n");

#ifdef DO_PWRLVL
	epson_debugprint("epson_close: lowering power level\n");
	usb_set_device_pwrlvl0(softstate->dip);
#endif

	softstate->openstate=0;

   	return (0);
}

static int epson_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
		    cred_t *credp, int *rvalp){
   	epson_debugprint("epson: ioctl called\n");
   	return (ENXIO);
}

static int epson_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct epson_softstate *softstate;
   	instance=ddi_get_instance(dip);

	epson_debugprint("epson: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(epson_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"epson_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	epson_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}


