/*
 * The sole purpose of this "driver"
 * is to fake attaching to any USB device you choose, and
 * print out the endpoint info.
 *
 * Unfortunatley, it wont attach right now. sigh.
 *
 * There doesnt currently (oct, 2001) seem to be a way to get the info
 * through prtconf
 *
 *    Philip Brown
 */

#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/usb/usba.h>
/* Durnit, I want to use -N but dont know the exact magic */
char _depends_on[] = "misc/usba"; 

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>

#define DEVNODE_NAME	"usbprobe"	
#define DEVNODE_TYPE	DDI_NT_SERIAL /* see sunddi.h */


/**********************************************************************/

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


static	int	usbdummy_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	usbdummy_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


/* "char/block operations" */
static struct cb_ops	usbdummy_cb_ops = {
	nodev,			/* open() */
	nodev,			/* close() */
	nodev,			/* strategy()		*/
	nodev,			/* print routine	*/
	nodev,			/* no dump routine	*/
	nodev,			/* read() */
	nodev,			/* write() */
	nodev,			/* generic ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP |D_64BIT,	/* safe for multi-thread/multi-processor */
				/* also theoretically 64bit safe */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops usbdummy_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	nulldev,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	nulldev,		/* devo_probe */
	usbdummy_attach,		/* devo_attach */
	usbdummy_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&usbdummy_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"USB probing driver",
	&usbdummy_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   usbdummy_global_mutex;


/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;

	va_start(args);
	fmt = va_arg(args, char *);
	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}



/*Like the usbdummy_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See usbdummy_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	debugprint("usbdummy: _init called\n");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&usbdummy_global_mutex, "usbdummy: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	if ((ret = mod_install(&modlinkage)) != 0) {
		cmn_err(CE_NOTE, "usbdummy: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	debugprint("usbdummy: _init returning SUCCESS, supposedly\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	debugprint("usbdummy: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	cmn_err(CE_NOTE, "usbdummy: _fini called\n");
	
	ret = mod_remove(&modlinkage);

	return (ret);
}

/************************************************************/



/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in usbdummy_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  usbdummy_detach()
 */
static int usbdummy_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance, endpoint_count,tmpint;
	size_t usb_raw_len;
	uchar_t *usb_rawconf;
	usb_config_descr_t usb_config;
	usb_endpoint_descr_t tmp_endpoint;
	int usb_interfacenum;
	usb_interface_descr_t usb_interface_d;



	debugprint("usbdummy: usbdummy_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "usbdummy: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "usbdummy: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "usbdummy: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	debugprint("usbdummy: attach for instance %d\n", instance);



	usb_rawconf = usb_get_raw_config_data(dip,&usb_raw_len);

	/* Fill out "configuration" descriptor */
	if(usb_parse_configuration_descr(usb_rawconf, usb_raw_len,
					&usb_config,
					USB_CONF_DESCR_SIZE) !=
						USB_CONF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_configuration_desc failed\n");
		return DDI_FAILURE;
	}
	debugprint("config num is %d\n",
			 usb_config.bConfigurationValue);
	usb_interfacenum=usb_get_interface_number(dip);
	debugprint("USB interface num is %d\n",
				usb_interfacenum);

	/* If interfacenum==-1, set it to 0: we are responsible for full dev*/
	if(usb_interfacenum==-1){
		usb_interfacenum=0;
	}

	/* Fill out "interface" descriptor */
	if(usb_parse_interface_descr(usb_rawconf,usb_raw_len,
				    usb_interfacenum,
				    0,
				    &usb_interface_d,
				    USB_IF_DESCR_SIZE) != USB_IF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_interface_desc failed");
		return DDI_FAILURE;
	}

	endpoint_count=usb_interface_d.bNumEndpoints;

	for(tmpint=0;tmpint<endpoint_count; tmpint++){
		if(usb_parse_endpoint_descr(usb_rawconf, usb_raw_len,
					usb_interfacenum,
					0,/* alt interfacenum*/
					tmpint,/*endpoint index num*/
					&tmp_endpoint,
					USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
		{
			cmn_err(CE_WARN, "could not parse endpoint %d",
				tmpint);
			continue;
		}
		debugprint("endpoint %d: direction = %s, attribs=%d\n",
		       tmp_endpoint.bEndpointAddress,
		       (tmp_endpoint.bEndpointAddress&USB_EPT_DIR_IN)>0 ?
			   "IN" : "OUT",
		       tmp_endpoint.bmAttributes);
		switch(tmp_endpoint.bmAttributes & USB_EPT_ATTR_MASK){
			case USB_EPT_ATTR_CONTROL:
			debugprint("endpoint is a CONTROL endpoint\n");
			break;
			case USB_EPT_ATTR_ISOCH:
			debugprint("endpoint is an isochronous endpoint\n");
			break;
			case USB_EPT_ATTR_BULK:
			debugprint("endpoint is a bulk transfer endpoint\n");
			break;
			case USB_EPT_ATTR_INTR:
			debugprint("endpoint is an interrupt endpoint\n");
			break;

		}
		
		
	}

	debugprint("INTERFACE class=%x, subclass=%x\n",
			 usb_interface_d.bInterfaceClass,
 			 usb_interface_d.bInterfaceSubClass);


	/* We never really want to attach to dev */
	return DDI_FAILURE;

}


static int usbdummy_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
   	instance=ddi_get_instance(dip);

	debugprint("epson: detach for instance %d. nothing to do\n",
			    instance);

	return (DDI_SUCCESS);
		     
}


